@@ -4,6 +4,13 @@ A tool for extracting useful information from youtube video's, like comments, or
 
 Author: Willem Hengeveld <itsme@xs4all.nl>
 
+todo:
+extract 4 different dictionaries from the html page.
+-- ytInitialPlayerResponse
+-- ytcfg.set()
+-- ytplayer.web_player_context_config
+-- ytInitialData 
+
 """
 
 import urllib.request
@@ -15,12 +22,39 @@ import sys
 import html
 import datetime
 from collections import defaultdict
-
 from xml.parsers.expat import ParserCreate
 
 import http.client
 
 
+def load_socks_proxy(proxyarg):
+    m = re.match(r'(?:(\w+)://)?(\S+):(\d+)', proxyarg)
+    if not m:
+        return
+    method, host, port = m.groups()
+    port = int(port)
+
+    if not method or not method.startswith('socks'):
+        return
+
+    import socks
+    socks.setdefaultproxy(socks.SOCKS4 if method.startswith('socks4') else socks.SOCKS5, host, port)
+    def create_connection(address, timeout=None, source_address=None):
+        sock = socks.socksocket()
+        sock.connect(address)
+        return sock
+    import socket
+    socket.create_connection = create_connection
+    socket.socket = socks.socksocket
+
+
+def decode_proxy(proxyarg):
+    if m:= re.match(r'(?:(\w+)://)?(\S+):(\d+)', proxyarg):
+        method, host, port = m.groups()
+        port = int(port)
+        if not method or method.startswith('http'):
+            return { 'http': proxyarg, 'https': proxyarg }
+
 def cvdate(txt):
     """
     Convert a string with a date in ymd format to a date object.
@@ -68,6 +102,16 @@ def getitem(d, *path):
             return
     return d
 
+def extracttext(entry):
+    return entry.get("simpleText") or "".join(r.get('text') for r in entry.get("runs"))
+
+
+def getcontinuation(p):
+    p = getitem(p, "continuations", 0, "nextContinuationData")
+    if not p:
+        return
+    return p["continuation"], p["clickTrackingParams"]
+
 
 class Youtube:
     """
@@ -80,11 +124,15 @@ class Youtube:
         self.args = args
         cj = http.cookiejar.CookieJar()
         handlers = [urllib.request.HTTPCookieProcessor(cj)]
+        if args.proxy:
+            proxies = decode_proxy(args.proxy)
+            if proxies:
+                handlers.append(urllib.request.ProxyHandler(proxies))
         if args.debug:
             handlers.append(urllib.request.HTTPSHandler(debuglevel=1))
         self.opener = urllib.request.build_opener(*handlers)
         self.innertubeapikey = "AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8"
-        self.clientversion =  "2.20200603.01.00"
+        self.clientversion =  "2.20210111.08.00"
 
     def httpreq(self, url, data=None):
         """
@@ -105,7 +153,13 @@ class Youtube:
             kwargs["data"] = data
 
         response = self.opener.open(req, **kwargs)
-        return response.read()
+        try:
+            page = response.read()
+        except http.client.IncompleteRead as e:
+            page = e.partial
+            print("EXCEPTION FOUND: http.client.IncompleteRead")
+            pass
+        return page
 
     def getcomments(self, contclick, xsrf):
         """
@@ -124,11 +178,29 @@ class Youtube:
         postdata = urllib.parse.urlencode({ "session_token":xsrf })
         return self.httpreq(url + "?" + urllib.parse.urlencode(query), postdata.encode('ascii') )
 
-    def getchat(self, cont, offset):
+    def getchat(self, cont, live=False):
         """
         Returns chat for the specified continuation parameter.
         """
-        url = "https://www.youtube.com/live_chat_replay/get_live_chat_replay"
+        if live:
+            url = "https://www.youtube.com/live_chat"
+        else:
+            url = "https://www.youtube.com/live_chat_replay"
+        query = {
+            "pbj": 1,
+            "continuation": cont,
+        }
+
+        return self.httpreq(url + "?" + urllib.parse.urlencode(query))
+
+    def getchat2(self, cont, offset, live=False):
+        """
+        Returns chat for the specified continuation parameter.
+        """
+        if live:
+            url = "https://www.youtube.com/youtubei/v1/live_chat_replay/get_live_chat"
+        else:
+            url = "https://www.youtube.com/youtubei/v1/live_chat_replay/get_live_chat_replay"
         query = {
             "pbj": 1,
             "continuation": cont,
@@ -230,16 +302,16 @@ class Youtube:
 
 class LivechatReader:
     """
-    class which can recursively print comments
+    class reads a livechat or livechat replay.
     """
-    def __init__(self, args, yt, cfg):
+    def __init__(self, args, yt, cfg, live=False):
         self.args = args
         self.yt = yt
+        self.live = live
         self.cont = self.getchatinfo(cfg)
 
     def getcontinuation(self, p):
-        p = getitem(p, "continuations", 0, "reloadContinuationData")
-        # or "liveChatReplayContinuationData" or "playerSeekContinuationData"
+        p = getitem(p, "continuations", 0, "liveChatReplayContinuationData" if self.live else "reloadContinuationData")
         if not p:
             return
         return p["continuation"]
@@ -249,7 +321,7 @@ class LivechatReader:
         Find the base parameters for querying the video's comments.
 
         """
-        item = getitem(cfg, ("playerResponse",), "response", "contents", "twoColumnWatchNextResults", "conversationBar", "liveChatRenderer")
+        item = getitem(cfg, ("response",), "response", "contents", "twoColumnWatchNextResults", "conversationBar", "liveChatRenderer")
         if not item:
             return
         return self.getcontinuation(item)
@@ -260,11 +332,15 @@ class LivechatReader:
             return
         ms = 0
         while True:
-            cmtjson = self.yt.getchat(self.cont, ms)
+            #cmtjson = self.yt.getchat2(self.cont, ms, self.live)
+            cmtjson = self.yt.getchat(self.cont, self.live)
             if self.args.debug:
                 print("============ chat req")
                 print(cmtjson.decode('utf-8'))
                 print()
+            if cmtjson.startswith(b"<!DOCTYPE"):
+                print("not json")
+                break
 
             js = json.loads(cmtjson)
 
@@ -272,14 +348,14 @@ class LivechatReader:
             if newms==ms:
                 break
 
-            for author, time, runs in cmtlist:
+            for author, time, comment in cmtlist:
                 print("--->", time, author)
-                self.printtextrun(runs)
+                print(extracttext(comment))
 
             ms = newms
 
     def extractchat(self, js):
-        actions = getitem(js, "response", "continuationContents", "liveChatContinuation", "actions")
+        actions = getitem(js, ("response",), "response", "continuationContents", "liveChatContinuation", "actions")
 
         cmtlist = []
         ms = None
@@ -290,7 +366,7 @@ class LivechatReader:
 
             item = getitem(replay, ("addChatItemAction",), "addChatItemAction", "item", "liveChatTextMessageRenderer")
             if item:
-                msg = getitem(item, "message", "runs")
+                msg = getitem(item, "message")
                 author = getitem(item, "authorName", "simpleText")
                 time = getitem(item, "timestampText", "simpleText")
 
@@ -298,11 +374,6 @@ class LivechatReader:
 
         return cmtlist, ms
 
-    def printtextrun(self, runs):
-        for r in runs:
-            print(r.get('text'), end="")
-        print()
-
 
 
 class CommentReader:
@@ -314,11 +385,6 @@ class CommentReader:
         self.yt = yt
         self.contclick, self.xsrf = self.getcommentinfo(cfg)
 
-    def printtextrun(self, runs):
-        for r in runs:
-            print(r.get('text'), end="")
-        print()
-
     def recursecomments(self, cc=None, level=0):
         if not cc:
             cc = self.contclick
@@ -333,25 +399,22 @@ class CommentReader:
 
             cmtlist, cc = self.extractcomments(js) 
 
-            for author, runs, subcc in cmtlist:
-                print("---" * (level+1) + ">", author)
-                self.printtextrun(runs)
+            for author, when, comment, likes, replies, subcc in cmtlist:
+                if self.args.verbose:
+                    print("---" * (level+1) + ">", "%s ; %s ; %s likes ; %s replies" % (author, when, likes, replies))
+                else:
+                    print("---" * (level+1) + ">", author)
+                print(extracttext(comment))
                 if subcc:
                     self.recursecomments(subcc, level+1)
 
-    def getcontinuation(self, p):
-        p = getitem(p, "continuations", 0, "nextContinuationData")
-        if not p:
-            return
-        return p["continuation"], p["clickTrackingParams"]
-
     def getcommentinfo(self, cfg):
         """
         Find the base parameters for querying the video's comments.
 
         """
         item = getitem(cfg, ("response",), "response", "contents", "twoColumnWatchNextResults", "results", "results", "contents")
-        cont = self.getcontinuation(getitem(item, ("itemSectionRenderer",), "itemSectionRenderer")) 
+        cont = getcontinuation(getitem(item, ("itemSectionRenderer",), "itemSectionRenderer")) 
         xsrf = getitem(cfg, ("response",), "xsrf_token")
 
         return cont, xsrf
@@ -373,14 +436,17 @@ class CommentReader:
             r = r["replies"]
 
         author = getitem(c,  "authorText", "simpleText")
-        content = getitem(c,  "contentText", "runs")
+        content = getitem(c,  "contentText")
+        likes = getitem(c, "likeCount")
+        nrreplies = getitem(c, "replyCount")
+        when = extracttext(getitem(c,  "publishedTimeText"))
         replies = getitem(r,  "commentRepliesRenderer")
         if replies:
-            cont = self.getcontinuation(replies)
+            cont = getcontinuation(replies)
         else:
             cont = None
 
-        return author, content, cont
+        return author, when, content, int(likes or 0), int(nrreplies or 0), cont
 
     def extractcomments(self, js):
         """
@@ -401,7 +467,7 @@ class CommentReader:
 
         # header.commentsHeaderRenderer -> commentsCount  at same level as 'contents'
 
-        return cmtlist, self.getcontinuation(p)
+        return cmtlist, getcontinuation(p)
 
 
 class SearchReader:
@@ -410,12 +476,6 @@ class SearchReader:
         self.yt = yt
         self.cfg = cfg
 
-    def extractruns(self, runs):
-        text = []
-        for r in runs:
-            text.append(r.get('text'))
-        return "".join(text)
-
     def getresults(self, js):
         ct = getitem(js, "contents", "twoColumnSearchResultsRenderer", "primaryContents", "sectionListRenderer", "contents")
         if not ct:
@@ -433,14 +493,14 @@ class SearchReader:
                 if video := item.get("videoRenderer"):
                     vid = getitem(video, "videoId")
                     pub = getitem(video, "publishedTimeText", "simpleText")
-                    title = getitem(video, "title", "runs")
+                    title = getitem(video, "title")
                     # title -> runs
                     # descriptionSnippet -> runs
                     # publishedTimeText -> simpleText
                     # lengthText -> simpleText
                     # viewCountText -> simpleText
                     # ownerText -> runs
-                    print("%s - %s" % (vid, self.extractruns(title)))
+                    print("%s - %s" % (vid, extracttext(title)))
                 elif chan := item.get("channelRenderer"):
                     cid = getitem(chan, "channelId")
                     title = getitem(chan, "title", "simpleText")
@@ -559,7 +619,7 @@ class SubtitleReader:
             print("========== timedtext xml")
             print(ttxml.decode('utf-8'))
             print()
-        tt = self.extracttext(ttxml)
+        tt = self.extractxmltext(ttxml)
 
         if self.args.srt:
             self.output_srt(tt)
@@ -596,7 +656,7 @@ class SubtitleReader:
         """
         return html.unescape(re.sub(r'</?font[^>]*>', '', htmltext))
 
-    def extracttext(self, xml):
+    def extractxmltext(self, xml):
         """
         Returns a list of tuples: time, endtime, text
         """
@@ -640,6 +700,10 @@ class PlaylistReader:
         self.cfg = cfg
 
     def output(self):
+        # ==== [                  'playlistVideoRenderer', 1, 'contents', 'playlistVideoListRenderer', 0, 'contents', 'itemSectionRenderer', 0, 'contents', 'sectionListRenderer', 'content', 'tabRenderer', 0, 'tabs', 'twoColumnBrowseResultsRenderer', 'contents', 'response', 1]
+        # ==== ['gridVideoRenderer', 1, 'items', 'horizontalListRenderer', 'content', 'shelfRenderer', 0, 
+        #                     'contents', 'itemSectionRenderer', 1, 'contents', 'sectionListRenderer', 'content', 'tabRenderer', 0, 
+        #                     'tabs', 'twoColumnBrowseResultsRenderer', 'contents', 'response', 1]
         playlist = getitem(self.cfg, ("response",), "response", "contents", "twoColumnWatchNextResults", "playlist")
         if playlist:
             print("Title: %s" % getitem(playlist, "playlist", "title"))
@@ -653,13 +717,28 @@ class PlaylistReader:
                     print("%s - %s" % (vid, title))
             return
         tabs = getitem(self.cfg, ("response",), "response", "contents", "twoColumnBrowseResultsRenderer", "tabs", 0, "tabRenderer", "content")
-        playlist = getitem(tabs, "sectionListRenderer", "contents", 0, "itemSectionRenderer", "contents", 0, "playlistVideoListRenderer")
+        ct1 = getitem(tabs, "sectionListRenderer", "contents", 0, "itemSectionRenderer", "contents", 0)
+        playlist = getitem(ct1, "playlistVideoListRenderer")
+        list_tag = "contents"
+        entry_tag = "playlistVideoRenderer"
+        if not playlist:
+            playlist = getitem(ctl, "shelfRenderer", "content", 'horizontalListRenderer')
+            list_tag = "items"
+            entry_tag = "gridVideoRenderer"
         if playlist:
-            for entry in playlist["contents"]:
-                vid = getitem(entry, "playlistVideoRenderer", "videoId")
-                title = getitem(entry, "playlistVideoRenderer", "title", "simpleText")
-                print("%s - %s" % (vid, title))
-            cont = self.getcontinuation(playlist)
+            cont = None
+            for entry in playlist[list_tag]:
+                vid = getitem(entry, entry_tag, "videoId")
+                title = getitem(entry, entry_tag, "title")
+                if vid and title:
+                    print("%s - %s" % (vid, extracttext(title)))
+                c = getitem(entry, "continuationItemRenderer", "continuationEndpoint", "continuationCommand", "token")
+                if c:
+                    cl = getitem(entry, "continuationItemRenderer", "continuationEndpoint", "clickTrackingParams")
+                    cont = c, cl
+
+            if not cont:
+                cont = getcontinuation(playlist)
             while cont:
                 browsejson = self.yt.browse(cont)
                 if self.args.debug:
@@ -669,29 +748,36 @@ class PlaylistReader:
 
                 js = json.loads(browsejson)
 
+                cont = None
                 playlist = getitem(js, ("response",), "response", "continuationContents", "gridContinuation")
                 if playlist:
                     for entry in getitem(playlist, "items"):
                         vid = getitem(entry, "gridVideoRenderer", "videoId")
-                        title = getitem(entry, "gridVideoRenderer", "title", "simpleText")
-                        print("%s - %s" % (vid, title))
+                        title = getitem(entry, "gridVideoRenderer", "title")
+                        print("%s - %s" % (vid, extracttext(title)))
                 playlist = getitem(js, ("response",), "response", "continuationContents", "playlistVideoListContinuation")
+                item_tag = "contents"
+                if not playlist:
+                    playlist = getitem(js, ("response",), "response", "onResponseReceivedActions", 0, "appendContinuationItemsAction", )
+                    item_tag = "continuationItems"
                 if playlist:
-                    for entry in getitem(playlist, "contents"):
+                    for entry in getitem(playlist, item_tag):
                         vid = getitem(entry, "playlistVideoRenderer", "videoId")
-                        title = getitem(entry, "playlistVideoRenderer", "title", "simpleText")
-                        print("%s - %s" % (vid, title))
+                        title = getitem(entry, "playlistVideoRenderer", "title")
+                        if vid and title:
+                            print("%s - %s" % (vid, extracttext(title)))
+                        c = getitem(entry, "continuationItemRenderer", "continuationEndpoint", "continuationCommand", "token")
+                        if c:
+                            cl = getitem(entry, "continuationItemRenderer", "continuationEndpoint", "clickTrackingParams")
+                            cont = c, cl
 
-                cont = self.getcontinuation(playlist)
+                if not playlist:
+                    break
+                if not cont:
+                    cont = getcontinuation(playlist)
 
             return
 
-    def getcontinuation(self, p):
-        p = getitem(p, "continuations", 0, "nextContinuationData")
-        if not p:
-            return
-        return p["continuation"], p["clickTrackingParams"]
-
 
 def parse_youtube_link(url):
     """
@@ -704,6 +790,7 @@ def parse_youtube_link(url):
     (?:www.)?youtube.com...
 
     /channel/<channelid>
+    /c/<channelname>
     /playlist?list=<listid>
     /watch?v=<videoid> [&t=pos] [&list=<listid>]
     /watch/<videoid>
@@ -713,7 +800,6 @@ def parse_youtube_link(url):
     /watch_videos?video_ids=<videoid>,<videoid>,...
     /results?search_query=...
     """
-
     m = re.match(r'^(?:https?://)?(?:www\.)?(?:(?:youtu\.be|youtube\.com)/)?(.*)', url)
     if not m:
         raise Exception("youtube link not matched")
@@ -728,6 +814,8 @@ def parse_youtube_link(url):
             idtype = 'video'
         elif idtype in ('channel'):
             idtype = 'channel'
+        elif idtype in ('c'):
+            idtype = 'channelname'
         elif idtype in ('playlist'):
             idtype = 'playlist'
         else:
@@ -779,9 +867,20 @@ def parse_youtube_link(url):
         raise Exception("unknown id")
 
 def channelurl_from_userpage(cfg):
-    return getitem(cfg, ("playerResponse",), "response", "metadata", "channelMetadataRenderer", "channelUrl")
+    return getitem(cfg, ("response",), "response", "metadata", "channelMetadataRenderer", "channelUrl")
+
+def check_error(cfg):
+    status = getitem(cfg, ("playerResponse",), "playerResponse", "playabilityStatus")
+    if not status:
+        return
+    if status["status"] == "ERROR":
+        print(status["reason"])
+        return True
 
 def main():
+    import io
+    sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding='utf8')
+
     import argparse
     parser = argparse.ArgumentParser(description='Extract Youtube comments')
     parser.add_argument('--debug', '-d', action='store_true', help='print all intermediate steps')
@@ -793,14 +892,17 @@ def main():
     parser.add_argument('--info', '-i', action='store_true', help='Print video info')
     parser.add_argument('--srt', action='store_true', help='Output subtitles in .srt format.')
     parser.add_argument('--query', '-q', action='store_true', help='List videos matching the specified query')
-    parser.add_argument('--livechat', action='store_true', help='Print chat contents')
+    parser.add_argument('--livechat', action='store_true', help='Follow livechat contents')
+    parser.add_argument('--replay', action='store_true', help='Print livechat replay')
+    parser.add_argument('--proxy', type=str, help='Specify a proxy to use.')
     parser.add_argument('ytids', nargs='+', type=str, help='One or more Youtube URLs, or IDs, or a query')
     args = parser.parse_args()
 
+    if args.proxy and args.proxy.startswith('socks'):
+        load_socks_proxy(args.proxy)
+
     yt = Youtube(args)
 
-
-
     for url in args.ytids:
         if len(args.ytids) > 1:
             print("==>", url, "<==")
@@ -825,6 +927,8 @@ def main():
                 url = "https://www.youtube.com/results?" + urllib.parse.urlencode({"search_query": idvalue})
 
             cfg = yt.getpageinfo(url)
+            if check_error(cfg):
+                continue
 
             if idtype=='username':
                 url = channelurl_from_userpage(cfg)
@@ -837,8 +941,8 @@ def main():
             if args.subtitles and idtype=='video':
                 txt = SubtitleReader(args, yt, cfg)
                 txt.output()
-            if args.livechat and idtype=='video':
-                txt = LivechatReader(args, yt, cfg)
+            if (args.replay or args.livechat) and idtype=='video':
+                txt = LivechatReader(args, yt, cfg, live=args.livechat)
                 txt.recursechat()
             if args.playlist and idtype=='playlist':
                 lst = PlaylistReader(args, yt, cfg)
